// crp2csv - create a CSV file for a height profile graph
// from a fragment of a CicloTour CRP file
// (C) 2003 Frank Paehlke <frank@paehlke.de>
// $Id: crp2csv.cc,v 1.4 2004/01/29 09:27:34 frank Exp $

// This program inputs a CRP file (or a fragment thereof), smoothes the
// data points as necessary and outputs two files:
//
// 1. An XML file containing meta-information like the total length,
//    height difference, maximum slope etc.
// 2. A CSV file containing the following tab-separated data rows:
//    time, orig. distance, orig. height, heart rate, temperature,
//    corrected distance, corrected height, speed, vertical speed,
//    slope, comment

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "config.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <getopt.h>
#include "globals.h"
#include "crpfile.h"

using namespace std;

// global variables

struct _globals globals;

// command-line options processing

struct option longopts[] = 
{
    {"help", no_argument, NULL, 'h'},
    {"version", no_argument, NULL, 'V'},
    {"verbose", no_argument, NULL, 'v'},
    {"nosmooth", no_argument, NULL, 'n'},
    {"xml", required_argument, NULL, 'x'},
    {"startheight", required_argument, NULL, 's'},
    {"endheight", required_argument, NULL, 'e'},
    {"addheight", required_argument, NULL, 'a'},
    {"lscale", required_argument, NULL, 'l'},
    {0,0,0,0}
};

void usage (int argc, char ** argv)
{
    cout << "Usage: " << argv[0] << " [OPTION...] [infile [outfile]]\n\n"
	"-h, --help      display this help and exit\n"
	"-V, --version   display version information and exit\n"
	"-v, --verbose   enable verbose output\n"
	"-n, --nosmooth  don't smooth distance and height values\n"
	"-x STRING, --xml STRING\n"
	"                set filename for XML output\n"
	"-s INT, --startheight INT\n"
	"                set start height [m] for profile scaling\n"
	"-e INT, --endheight INT\n"
	"                set end height [m] for profile scaling\n"
	"-a INT, --addheight INT\n"
	"                set height offset [m] for profile scaling\n"
	"                (cannot be combined with -s and -e)\n"
	"-l FLOAT, --lscale FLOAT\n"
	"                set length scaling factor\n";
}

void version (int argc, char ** argv)
{
    cout << argv[0] << " (" << PACKAGE << ") " << VERSION << endl;
}

int
main (int argc, char *argv[] )
{
    int opt;
    // set defaults for global variables
    globals.verbose = false;
    globals.nosmooth = false;
    globals.use_startheight = globals.use_endheight = false;
    globals.hoffset = 0;
    globals.lscale = 1.0;
    // read command line parameters
    ostream* outfile = &cout;
    string xmlfilename = "";
    istream* infile = &cin;

    while ((opt = getopt_long (argc, argv, "hVvnx:s:e:a:l:",
			       longopts, NULL)) != EOF) {
	switch (opt) {
	    case 'h':
		usage (argc, argv);
		exit(0);
	    case 'V':
                version (argc, argv);
                exit(0);
	    case 'v':
                globals.verbose = true;
                break;
	    case 'n':
                globals.nosmooth = true;
                break;
	    case 'x':
		xmlfilename = optarg;
		if (globals.verbose)
		    cerr << "writing XML file " << xmlfilename << endl;
                break;
	    case 's':
		if (globals.hoffset) {
		    cerr << "Error: cannot use -s together with -a" << endl;
		    exit(1);
		}
		globals.use_startheight = true;
		{
		    istringstream is (optarg);
		    is >> globals.startheight;
		}
		if (globals.verbose)
		    cerr << "new start height: " << globals.startheight
			 << endl;
                break;
	    case 'e':
		if (globals.hoffset) {
		    cerr << "Error: cannot use -e together with -a" << endl;
		    exit(1);
		}
		globals.use_endheight = true;
		{
		    istringstream is (optarg);
		    is >> globals.endheight;
		}
		if (globals.verbose)
		    cerr << "new end height: " << globals.endheight << endl;
		break;
	    case 'a':
		if (globals.use_startheight || globals.use_endheight) {
		    cerr << "Error: cannot use -a together with -s or -e"
			 << endl;
		    exit(1);
		}
		{
		    istringstream is (optarg);
		    is >> globals.hoffset;
		}
		if (globals.verbose)
		    cerr << "height offset: " << globals.hoffset << endl;
		break;
	    case 'l':
	    {
		istringstream is (optarg);
		is >> globals.lscale;
	    }
	    if (globals.verbose)
		cerr << "setting length scaling factor to "
		     << globals.lscale << endl;
	    break;
	} // switch (opt)
    } // while (opt)

    // input file name given as additional argument?
    if (optind == argc) {
	if (globals.verbose)
	    cerr << "reading from standard input\n";
    }
    else {
	string ifname = argv[optind];
	if (ifname == "-") {
	    if (globals.verbose)
		cerr << "reading from standard input\n";
	}
	else {
	    if (globals.verbose)
		cerr << "reading from file " << ifname << endl;
	    if (!(infile = new ifstream (ifname.c_str()))) {
		cerr << "Error: can't open input file" << endl;
		exit(1);
	    }
	}
	optind++;
    }

    // output file name given as additional argument?
    if (optind == argc) {
	if (globals.verbose)
	    cerr << "writing to standard output\n";
    }
    else {
	string ofname = argv[optind];
	if (ofname == "-") {
	    if (globals.verbose)
		cerr << "writing to standard output\n";
	}
	else {
	    if (globals.verbose)
		cerr << "writing output to file " << ofname << endl;
	    if (!(outfile = new ofstream (ofname.c_str()))) {
		cerr << "Error: can't open output file" << endl;
		exit(1);
	    }
	}
	optind++;
    }

    if (++optind < argc)
	cerr << "Warning: ignoring extra argument(s)" << endl;
    
    // Initialize CRP file

    CRPFile crp;
    if (!globals.nosmooth) {
	crp.set_smooth(new min_d2_smooth);
    }

    // read input file
    
    try {
	(*infile) >> crp;
    }
    catch (CRPFile::MalformedInputException& ex) {
	cerr << "Error: Malformed input" << endl;
	cerr.flush();
	exit(1);
    }
    if (infile != &cin)
	delete infile;
    
    // dump the file
    
    crp.writeCSV (*outfile);
    if (outfile != &cout)
	delete outfile;

    // write XML file if desired

    if (!xmlfilename.empty()) {
	ofstream *xmlfile = new ofstream (xmlfilename.c_str());
	if (!xmlfile) {
	    cerr << "Error: can't open XML file" << endl;
	    exit(1);
	}
	else {
	    crp.writeXML (*xmlfile);
	    delete xmlfile;
	}
    }
    
    exit(0);
}
