// Smoothing of data vectors
// (C) 2002 Frank Paehlke <frank@paehlke.de>
// $Id: smooth.cc,v 1.3 2004/01/29 09:32:45 frank Exp $

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "smooth.h"

#include <cmath>

using namespace std;

// The target function:
// tf(X) = sum ((2*x[i] - x[i-1] - x[i+1])^2, i=1..n-1)

double min_d2_smooth::tf (const vector<double>& X) const throw()
{
    int n = X.size()-1;
    double res = 0;
    for (int i=1; i<n; i++) {
	double a = 2*X[i] - X[i-1] - X[i+1];
	res += a*a;
    }
    return res;
} 

// partial derivative of target function by X[i]

double min_d2_smooth::dtf(const vector<double>& X, int i) const
    throw(out_of_range)
{
    int n = X.size()-1;
    if ((i < 0) || (i > n))
	throw out_of_range("Index out of range");
    if (n < 2)
	return 0;
    else if (i == 0)
	return X[0] - 2*X[1] + X[2];
    else if (i == 1)
	return -2*X[0] + 5*X[1] - 4*X[2] + X[3];
    else if (i == n-1)
	return X[i-2] - 4*X[i-1] + 5*X[i] - 2*X[i+1];
    else if (i == n)
	return X[i-2] - 2*X[i-1] + X[i];
    else 
	return X[i-2] - 4*X[i-1] + 6*X[i] - 4*X[i+1] + X[i+2];
}

// 2nd partial derivative of target function by X[i]

double min_d2_smooth::d2tf(const vector<double>& X, int i) const
    throw(out_of_range)
{
    int n = X.size()-1;
    if ((i < 0) || (i > n))
	throw out_of_range("Index out of range");
    if (n < 2)
	return 0;
    else if ((i == 0) || (i == n))
	return 1;
    else if ((i == 1) || (i == n-1))
	return 5;
    else
	return 6;		
}

// minimize tf(X) under the condition (X[i]-X0[i] <= epsilon)

// tf(X) is minimized using an iterative algorithm until the
// maximum change to any X[i] between two subsequent iterations
// is smaller than "accuracy"

void min_d2_smooth::smooth_vector
(const vector<double>& X0, vector<double>& X, double epsilon) const throw()
{
    X = X0;
    int n = X0.size()-1;
    if (n < 2)
	return;
    double maxcorr; // maximum correction
    do {
	maxcorr = 0;
	for (int i=1; i<n; i++) {
	    // correct X[i]
	    double dtfx = dtf(X,i);
	    double d2tfx = d2tf(X,i);
	    // correction term
	    double Xnew = X[i] - dtfx/d2tfx;
	    // border condition
	    if (Xnew < X0[i] - epsilon)
		Xnew = X0[i] - epsilon;
	    else if (Xnew > X0[i] + epsilon)
		Xnew = X0[i] + epsilon;
	    // maximum correction?
	    double corr = fabs(Xnew - X[i]);
	    if (corr > maxcorr)
		maxcorr = corr;
	    X[i] = Xnew;
	} // for(i)
    } while (maxcorr >= accuracy);
}
